/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <QtGui/QMessageBox>
#include <boost/optional.hpp>

#include <wipal/pcap/find.hh>

#include "wscout_gui_find_dialog.hh"

namespace wscout
{

  namespace gui
  {

    /*-----------.
    | FindThread |
    `-----------*/

    void
    FindThread::run()
    {
      const pcapxx::match	m (bytes_, first_, last_);
      const_iterator		i = current_it_;
      unsigned			c = 0;
      size_t			last_id = (i == last_it_?
					     0 : i->pcap_header()->id);

      while (i != last_it_ and not (*i == m))
	{
	  if (canceled_)
	    return;
	  last_id = i->pcap_header()->id;
	  if (not (c++ % positionUpdateInterval))
	    positionChanged(last_id);
	  ++i;
	}
      positionChanged(last_id);
      current_it_ = i;
    }

    /*-----------.
    | FindDialog |
    `-----------*/

    FindDialog::FindDialog(QWidget* parent): Dialog (parent)
    {
      const QStyle&	s = *(qApp->style());
      QGridLayout&	l = layout();

      setupBytesRow(l);
      setupOffsetsRow(l);
      setupProgressBar(l);
      setupButtonsRow(l, s);
    }

    void
    FindDialog::restartClicked()
    {
      cancelThread();
      current_it_ = first_it_;
      findNextClicked();
    }

    void
    FindDialog::findNextClicked()
    {
      assert(not threadRunning());
      assert(validIterators());

      try
	{
	  thread_.reset(new FindThread (this,
					current_it_.get(), last_it_.get(),
					bytes_.text().toStdString(),
					first_.value(), last_.value()));
	}
      catch (const std::invalid_argument& e)
	{
	  QMessageBox::critical(this,
				tr("Invalid argument"), tr(e.what()) + '.');
	  return;
	}
      connect(thread_.get(), SIGNAL(finished()), this, SLOT(threadFinished()));
      connect(thread_.get(), SIGNAL(positionChanged(int)),
	      &pbar_, SLOT(setValue(int)));
      find_next_.setEnabled(false);
      cancel_.setEnabled(true);
      thread_->start();
    }

    void
    FindDialog::threadFinished()
    {
      assert(thread_.get());

      if (not thread_->canceled())
	{
	  if (current_it_ == last_it_)
	    {
	      notFound();
	      current_it_ = first_it_;
	    }
	  else
	    {
	      found(current_it_.get()->pcap_header()->id - 1);
	      if (++current_it_.get() == last_it_)
		current_it_ = first_it_;
	    }
	}
      find_next_.setEnabled(true);
      cancel_.setEnabled(false);
    }

    void
    FindDialog::setupBytesRow(QGridLayout& l)
    {
      bytes_.setToolTip(tr("<p style=\"margin-bottom: 3px;"
			   "           margin-left: 8px;"
			   "           margin-right: 8px;\">"
			   "  <b>Bytes to look for.</b>"
			   "</p>"
			   "<p style=\"margin-top: 3px;"
			   "           margin-left: 8px;"
			   "           margin-right: 8px;\">"
			   "  <em>e.g.</em><br />"
			   "  <tt>my-ssid</tt><br />"
			   "  <tt>0xFEA0</tt><br />"
			   "  <tt>0Xd0d0fabecc</tt><br />"
			   "  etc."
			   "</p>"));
      bytes_.setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);

      l.addWidget(makeLabel(tr("Bytes")),			1, 1);
      l.addWidget(&bytes_,					1, 2, 1, 4);
    }

    void
    FindDialog::setupOffsetsRow(QGridLayout& l)
    {
      first_.setMinimum(first_min);
      first_.setMaximum(first_max);
      first_.setValue(first_def);
      first_.setToolTip(tr("Ignore this amount of bytes at the beginning of "
			   "each frame."));

      last_.setMinimum(last_min);
      last_.setMaximum(last_max);
      last_.setValue(last_def);
      last_.setToolTip(tr("Do noot look past this amount of bytes "
			  "in each frame."));

      l.addWidget(makeLabel(tr("Between offsets")),		2, 1);
      l.addWidget(&first_,					2, 2);
      l.addWidget(makeLabel(tr("and"), Qt::AlignHCenter),	2, 3);
      l.addWidget(&last_,					2, 4);
    }

    void
    FindDialog::setupProgressBar(QGridLayout& l)
    {
      pbar_.setMaximum(1);
      pbar_.setMinimum(2);
      pbar_.setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);

      l.addWidget(&pbar_,					3, 1, 1, 4);
    }

    void
    FindDialog::setupButtonsRow(QGridLayout& l, const QStyle& s)
    {
      // Cancel.
      {
	const QIcon&	i = s.standardPixmap(QStyle::SP_DialogCancelButton);

	cancel_.setIcon(i);
	cancel_.setText(tr("&Cancel"));
	cancel_.setToolTip(tr("Cancel current search."));
	cancel_.setEnabled(false);
	connect(&cancel_, SIGNAL(clicked(bool)), this, SLOT(cancelThread()));
      }

      // Restart.
      {
	const QIcon&	i = s.standardPixmap(QStyle::SP_DialogResetButton);

	restart_.setIcon(i);
	restart_.setText(tr("&Restart"));
	restart_.setToolTip(tr("Restart search from the beginning."));
	connect(&restart_, SIGNAL(clicked(bool)), SLOT(restartClicked()));
      }

      // Find Next.
      {
	const QIcon&	i = s.standardPixmap(QStyle::SP_DialogApplyButton);

	find_next_.setIcon(i);
	find_next_.setText(tr("&Find next"));
	find_next_.setToolTip(tr("Find next occurrence."));
	find_next_.setDefault(true);
	connect(&find_next_, SIGNAL(clicked(bool)), SLOT(findNextClicked()));
      }

      // Close.
      {
	const QIcon&	i = s.standardPixmap(QStyle::SP_DialogCloseButton);

	close_.setIcon(i);
	close_.setText(tr("C&lose"));
	close_.setToolTip(tr("Close dialog."));
	close_.setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
	connect(&close_, SIGNAL(clicked(bool)), this, SLOT(cancelThread()));
	connect(&close_, SIGNAL(clicked(bool)), this, SLOT(close()));
      }

      l.addWidget(&cancel_,					4, 1);
      l.addWidget(&restart_,					4, 2);
      l.addWidget(&find_next_,					4, 3);
      l.addWidget(&close_,					4, 4);
    }

  } // End of namespace gui.

} // End of namespace wscout.

