/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

# include <wipal/tool/endianness.hh>

# include <sstream>

# include "wscout_gui_ieee802_11_beacon_header.hh"

namespace wscout
{

  namespace gui
  {

    IEEE80211BeaconHeader::IEEE80211BeaconHeader(QWidget* parent):
      HeaderBase<IEEE80211BeaconHeader, protocol::IEEE802_11_BEACON> (parent)
    {
      setUpdatesEnabled(false);

      {
	QPalette	p;

	p.setColor(QPalette::WindowText, QColor ("black"));
	p.setColor(QPalette::Window, QColor ("lightblue"));
	setForegroundRole(QPalette::WindowText);
	setBackgroundRole(QPalette::Window);
	setAutoFillBackground(true);
	setPalette(p);
      }

      {
	QFont bold, fixed ("Fixed");
	bold.setBold(true);

	ssid.setFont(bold);
	capability.setFont(fixed);
	timestamp.setFont(fixed);
	beaconInterval.setAlignment(Qt::AlignRight);
      }

      {
	QLayout* const	l = layout();
	l->addWidget(&ssid);
	l->addWidget(&capability);
	l->addWidget(&timestamp);
	l->addWidget(&beaconInterval);
      }

      setUpdatesEnabled(true);
    }

    IEEE80211BeaconHeader::next_header
    IEEE80211BeaconHeader::setup(const void*		header,
			      size_t			caplen,
			      const Trace&		,
			      const pcapxx::frame_descriptor&)
    {
      ssid.setText("SSID: ?");

      if (not header or caplen < sizeof (frame_body))
	{
	  timestamp.setText("TSF: ?");
	  beaconInterval.setText("bcn. inter.: ?");
	  capability.setText("????????????????");

	  return next_header (protocol::NONE, 0);
	}

      const frame_body*	body = static_cast<const frame_body*> (header);
      const bool	swap = tool::endian::need_swap(tool::endian::little,
						       false);
      setupTimestamp(*body);
      setupBeaconInterval(*body, swap);
      setupCapability(*body, swap);
      setupInfoElements(*body, caplen, swap);

      return next_header (protocol::NONE, 0);
    }

    void
    IEEE80211BeaconHeader::setupTimestamp(const frame_body& body)
    {
      std::ostringstream	os;

      os << body.timestamp;
      timestamp.setText(tr("TSF: %1 us").arg(os.str().c_str()));
    }

    void
    IEEE80211BeaconHeader::setupBeaconInterval(const frame_body& bdy, bool sw)
    {
      const unsigned	bi = tool::extract_short_u(bdy.beacon_interval, sw);

      beaconInterval.setText(tr("bcn. inter.: %1 TUs (%2 ms)")
			     .arg(bi)
			     .arg(bi * 1024.f / 1000.f));
    }

    void
    IEEE80211BeaconHeader::setupCapability(const frame_body& body, bool swap)
    {
      char	s[17] = "0123456789ABCDEF";
      unsigned	f = tool::extract_short_u(body.capability, swap);

      for (unsigned i = 0; i < sizeof (s) - 1; ++i)
	if (not (f & (1 << i)))
	  s[i] = '-';

      capability.setText(tr("%1").arg(s));
    }

    void
    IEEE80211BeaconHeader::setupInfoElements(const frame_body&	body,
					     size_t		caplen,
					     bool		swap)
    {
      const frame_info_element*	e =
	reinterpret_cast<const frame_info_element*> (body.info_elements);

      size_t	len = sizeof (frame_body);

      while (len < caplen)
	{
	  const frame_info_element* n = handleInfoElement(e, swap);
	  len += (reinterpret_cast<const uint8_t*> (n) -
		  reinterpret_cast<const uint8_t*> (e));
	  e = n;
	}
    }

    const IEEE80211BeaconHeader::frame_info_element*
    IEEE80211BeaconHeader::handleInfoElement(const frame_info_element* e, bool)
    {
      switch (e->id)
	{
	case 0: // SSID.
	  ssid.setText(std::string (reinterpret_cast<const char*> (e->info),
				    e->len).c_str());
	  break;
	}

      return reinterpret_cast<const frame_info_element*> (e->info + e->len);
    }


  } // End of namespace gui.

} // End of namespace wscout.
