/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <sstream>
#include <iomanip>
#include <cstring>
#include <memory>

extern "C"
{
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <arpa/inet.h>
}

#include "wscout_gui_ip_header.hh"

namespace wscout
{

  namespace gui
  {

    IPv4Header::next_header
    IPv4Header::setup(const void*			hdr,
		      size_t				caplen,
		      const Trace&			,
		      const pcapxx::frame_descriptor&	)
    {
      const iphdr* const hdr4 = static_cast<const struct iphdr*> (hdr);

      if (caplen < sizeof (*hdr4))
	{
	  srcDst_.setText(tr("truncated IPv4 datagram"));
	  ttl_.setText(tr("TTL: ??"));
	  fragInfo_.setText(tr("ID ??"));

	  return next_header (protocol::UNKNOWN, 0);
	}

      {
	const std::auto_ptr<char>
	  src (strdup(inet_ntoa(*reinterpret_cast<const struct in_addr*>
				(&hdr4->saddr))));
	const std::auto_ptr<char>
	       dst (strdup(inet_ntoa(*reinterpret_cast<const struct in_addr*>
				     (&hdr4->daddr))));
	srcDst_.setText(tr("%1 > %2").arg(src.get()).arg(dst.get()));
      }

      ttl_.setText(tr("TTL: %1").arg(hdr4->ttl));

      unsigned			fragData = htons(hdr4->frag_off);
      bool			dontFragment = fragData & 0x4000;
      bool			moreFragments = fragData & 0x2000;
      unsigned			fragOffset = fragData & 0x1FFF;
      std::ostringstream	fragInfo;

      fragInfo << "ID " << htons(hdr4->id);
      if (fragOffset)
	fragInfo << " (offset " << fragOffset << ", "
		 << (moreFragments ? "more fragments" : "last fragment");
      else
	fragInfo << (moreFragments?
		     " (first fragment":
		     " (not fragmented");
      fragInfo << (dontFragment ? ", don't fragment)" : ")");
      fragInfo_.setText(fragInfo.str().c_str());

      return next_header (protocol::UNKNOWN, 0);
    }

    IPv6Header::next_header
    IPv6Header::setup(const void*			hdr,
		      size_t				caplen,
		      const Trace&			,
		      const pcapxx::frame_descriptor&	)
    {
      const ip6_hdr* const hdr6 = static_cast<const struct ip6_hdr*> (hdr);

      if (caplen < sizeof (*hdr6))
	{
	  srcDst_.setText(tr("truncated IPv6 datagram"));
	  hopLimit_.setText(tr("hop limit: ??"));

	  return next_header (protocol::UNKNOWN, 0);
	}

      {
	const size_t	ipv6_addr_max_len = 40;
	char		src[ipv6_addr_max_len];
	char		dst[ipv6_addr_max_len];

	{
	  const void* const r =
	    inet_ntop(AF_INET6, static_cast<const void*> (&hdr6->ip6_src),
		      src, sizeof (src));
	  assert(r); (void) r; // Avoid unused variable warnings.
	}
	{
	  const void* const r =
	    inet_ntop(AF_INET6, static_cast<const void*> (&hdr6->ip6_dst),
		      dst, sizeof (dst));
	  assert(r); (void) r; // Avoid unused variable warnings.
	}

	srcDst_.setText(tr("%1 > %2").arg(src).arg(dst));
      }

      hopLimit_.setText(tr("hop limit: %1").arg(hdr6->ip6_hlim));

      return next_header (protocol::UNKNOWN, 0);
    }

  } // End of namespace gui.

} // End of namespace wscout.
