/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <QtGui/QResizeEvent>
#include <QtGui/QKeyEvent>
#include <QtGui/QToolTip>
#include <QtGui/QCursor>
#include <QtGui/QApplication>
#include <QtGui/QClipboard>

#include "wscout_gui_trace.hh"
#include "wscout_gui_packet.hh"

namespace wscout
{

  namespace gui
  {

    Trace::~Trace()
    {
      delete trace_;
    }

    void
    Trace::writeSelection(const QString& file)
    {
      typedef selection::list_of_ranges::const_iterator	const_iterator;

      assert(trace_);
      assert(not file.isNull());

      pcapxx::dumper		d = trace_->dumper(file.toStdString());
      const const_iterator	e = selection_.ranges().end();

      for (const_iterator i = selection_.ranges().begin(); i != e; ++i)
	{
	  typedef pcapxx::descriptor<>::const_iterator	packet_iterator;

	  packet_iterator	p = (*trace_)[i->first];

	  for (int j = i->first; j <= i->second; ++j)
	    d(*p++);
	}
    }

    void
    Trace::setPosition(int p)
    {
      if (p < 0)
	p = 0;
      if (p > lastPosition())
	p = lastPosition();

      pos_ = p;
      startUpdateTimer(set_position_update_delay);
      verticalScrollBar()->setValue(pos_);
    }

    void
    Trace::setSwapContent(bool b)
    {
      if (swap_content_ == b)
	return;

      swap_content_ = b;
      update();
    }

    void
    Trace::update()
    {
      update_timer_.stop();
      viewport()->setUpdatesEnabled(false);

      if (not trace_ or not trace_->size())
	{
	  BOOST_FOREACH(Packet* p, pkts_)
	    p->hide();
	  viewport()->setUpdatesEnabled(true);
	  return;
	}

      const pcapxx::descriptor<>::iterator	e = trace_->end();
      pcapxx::descriptor<>::iterator		i = (*trace_)[position()];

      unsigned	j = 0;
      Packet*	p;
      bool	need_resize = false;

      for (int y = trace_vmargin;
	   y + trace_vmargin < view_.height() and i != e;
	   y += p->minimumHeight() + packet_spacing)
	{
	  if (j < pkts_.size())
	    p = pkts_[j];
	  else
	    {
	      p = new Packet (&view_);
	      pkts_.push_back(p);
	    }

	  const int	old_width = p->minimumWidth();
	  const int	old_height = p->minimumHeight();

	  p->hide();
	  p->setup(*this, *i);
	  p->setSelected(selection_.contains(position() + j));
	  p->show();

	  if (p->minimumWidth() > old_width or
	      old_height != p->minimumHeight())
	    need_resize = true;

	  ++i;
	  ++j;
	}
      while (j < pkts_.size())
	pkts_[j++]->hide();

      if (need_resize)
	resizeView(view_.size(), false, i != e);
      viewport()->setUpdatesEnabled(true);
    }

    int
    Trace::packetAt(const QPoint& pos) const
    {
      const QWidget*	wgt = childAt(pos);
      const Packet*	pkt;
      do
	{
	  pkt = dynamic_cast<const Packet*>(wgt);
	  if (not pkt and wgt)
	    wgt = wgt->parentWidget();
	}
      while (not pkt and wgt);

      std::vector<Packet*>::const_iterator i =
	std::find(pkts_.begin(), pkts_.end(), pkt);

      return i == pkts_.end() ? -1 : position() + (i - pkts_.begin());
    }

    void
    Trace::keyPressEvent(QKeyEvent* ev)
    {
      QScrollBar& v = *verticalScrollBar();

      switch (ev->key())
	{
	case Qt::Key_Space:
	  setPosition(position() + v.pageStep());
	  break;
	case Qt::Key_Backspace:
	  setPosition(position() - v.pageStep());
	  break;
	default:
	  QAbstractScrollArea::keyPressEvent(ev);
	  break;
	}
    }

    void
    Trace::mouseMoveEvent(QMouseEvent* ev)
    {
      if (last_selected_ >= 0)
	{
	  const int p = packetAt(ev->pos());

	  if (p >= 0)
	    {
	      selection_ = selection::range (last_selected_, p);
	      update();
	    }
	}
    }

    void
    Trace::mousePressEvent(QMouseEvent* ev)
    {
      const int			p = packetAt(ev->pos());
      const Qt::MouseButton	b = ev->button();

      if (p >= 0 and b == Qt::LeftButton)
	{
	  switch (ev->modifiers())
	    {
	    case Qt::NoModifier:
	      selection_ = last_selected_ = p;
	      update();
	      break;

	    case Qt::ShiftModifier:
	      if (last_selected_ >= 0)
		{
		  selection_ = selection::range (last_selected_, p);
		  update();
		}
	      break;

	    case Qt::ControlModifier:
	      selection_.toggle(p);
	      last_selected_ = p;
	      update();
	      break;

	    default:
	      break;
	    }
	  last_left_clicked_ = p;
	  emit packetLeftClicked(last_left_clicked_);
	}
      else if (b == Qt::MidButton)
	{
	  bool		ok (false);
	  const int	pos = QApplication::clipboard()->text().toInt(&ok);

	  if (ok)
	    setPosition(pos - 1);
	}
    }

    void
    Trace::mouseDoubleClickEvent(QMouseEvent* ev)
    {
      const int			p = packetAt(ev->pos());
      const Qt::MouseButton	b = ev->button();

      if (p >= 0 and b == Qt::LeftButton)
	{
	  last_left_clicked_ = p;
	  emit packetDoubleLeftClicked(last_left_clicked_);
	}
    }

    void
    Trace::resizeEvent(QResizeEvent* ev)
    {
      viewport()->setUpdatesEnabled(false);
      resizeView(ev->size(), true, true);
      viewport()->setUpdatesEnabled(true);
    }

    void
    Trace::scrollContentsBy(int dx, int dy)
    {
      if (dy)
	{
	  QScrollBar&	b = *verticalScrollBar();

	  pos_ = b.value();
	  emit positionChanged(pos_);
	  startUpdateTimer(scroll_contents_by_update_delay);
	  if (b.isSliderDown())
	    {
	      const int	y = QCursor::pos().y();
	      const int	x = mapToGlobal(QPoint (width(), 0)).x();

	      QToolTip::showText(QPoint (x, y), positionString(), this);
	    }
	}
      if (dx)
	view_.move(-horizontalScrollBar()->value(), 0);
    }

    void
    Trace::resizeView(const QSize&	size,
		      bool		resize_pkts,
		      bool		update_if_needed)
    {
      const int view_height = size.height();
      int	view_width = 0;

      BOOST_FOREACH(Packet* p, pkts_)
	{
	  const int t = p->minimumWidth();
	  if (t > view_width)
	    view_width = t;
	}
      view_width = std::max(view_width + 2 * trace_hmargin, size.width());
      view_.resize(view_width, view_height);

      {
	const int	viewport_width = viewport()->width();
	QScrollBar&	b = *horizontalScrollBar();

	b.setRange(0, view_width - viewport_width);
	b.setPageStep(viewport_width);
      }

      const int packet_width = view_width - 2 * trace_hmargin;
      size_t	i;
      int	y;

      for (i = 0, y = trace_vmargin;
	   y + trace_vmargin < view_height and i < pkts_.size();
	   y += pkts_[i++]->minimumHeight() + packet_spacing)
	pkts_[i]->setGeometry(trace_hmargin, y,
			      packet_width, pkts_[i]->minimumHeight());
      if (resize_pkts)
	while (i < pkts_.size())
	  {
	    delete pkts_.back();
	    pkts_.pop_back();
	  }

      verticalScrollBar()->setPageStep(i - 1);
      if (update_if_needed and
	  y + trace_vmargin < view_height and trace_)
	update();
    }

  } // End of namespace gui.

} // End of namespace wscout.
