/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_FRAME_ITERATOR_HXX_
# define PCAP_FRAME_ITERATOR_HXX_

# include <fstream>
# include <stdexcept>
# include <memory>
# include <cassert>

# include "frame_iterator.hh"
# include <wipal/tool/exceptions.hh>

namespace pcapxx
{

  namespace internals
  {

    template <class B>
    frame_iterator<B>::frame_iterator(const iterable_type& desc, bool end):
      super_type (),
      desc_ (&desc)
    {
      if (not end)
	{
	  pos_ = sizeof (file_header);
	  count_ = 0;
	  increment();
	}
    }

    template <class B>
    frame_iterator<B>::frame_iterator(const iterable_type&	desc,
				      const std::streampos&	pos,
				      unsigned			count):
      super_type (),
      desc_ (&desc),
      pos_ (pos),
      count_ (count)
    {
      increment();
    }

    template <class B>
    frame_iterator<B>::frame_iterator(const frame_iterator& it):
      super_type (),
      desc_ (it.desc_)
    {
      if (it.value())
	{
	  pos_ = it.pos_;
	  count_ = it.count_;
	  this->value() = it.value();
	}
    }

    template <class B>
    frame_iterator<B>&
    frame_iterator<B>::operator = (const frame_iterator& rhs)
    {
      if (&rhs == this)
	return *this;

      desc_ = rhs.desc_;
      input_.release();
      if ((this->value() = rhs.value()))
	{
	  pos_ = rhs.pos_;
	  count_ = rhs.count_;
	}
      return *this;
    }

    template <class B>
    bool
    frame_iterator<B>::equal(const frame_iterator& rhs) const
    {
      if (not rhs.value())
	return not this->value();
      return (this->value() and
	      pos_ == rhs.pos_ and
	      desc_->file_name() == rhs.desc_->file_name());
    }

    template <class B>
    void
    frame_iterator<B>::increment()
    {
      try
	{
	  this->value() = frame_descriptor (stream(),
					    ++count_,
					    desc_->swapped(),
					    desc_->zone(),
					    &pos_);
	}
      catch (const tool::read_error&)
	{
	  if (stream().eof())
	    this->value() = boost::none_t ();
	  else
	    throw;
	}
    }

    template <class B>
    std::istream&
    frame_iterator<B>::stream()
    {
      if (not input_.get())
	{
	  input_.reset(new std::ifstream (desc_->file_name().c_str()));
	  if (input_->fail())
	    throw tool::file_error ("Open failed");
	  if (not input_->seekg(pos_, std::ios::beg))
	    throw tool::seek_error("Seek operation failed");
	}
      return *input_.get();
    }

  } // End of namespace pcapxx::internals.

} // End of namespace pcapxx.

#endif // ! PCAP_FRAME_ITERATOR_HXX_
