/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_STATS_HH_
# define WIFI_FRAME_STATS_HH_

# include <ostream>
# include <map>
# include <vector>

# include <boost/optional.hpp>
# include <boost/tuple/tuple.hpp>
# include <boost/tuple/tuple_comparison.hpp>

# include <wipal/wifi/frame/dissector/default_hooks.hh>
# include <wipal/pcap/frame_descriptor.hh>
# include <wipal/tool/microseconds.hh>
# include <wipal/wifi/frame/seqctl.hh>

namespace wifi
{

  namespace frame
  {

    namespace stats
    {

      /*---------------.
      | abstract_stats |
      `---------------*/

      struct abstract_stats
      {
	virtual			~abstract_stats();
	virtual std::ostream&	print(std::ostream& o) const = 0;
      };

      /*----------------.
      | simple_counters |
      `----------------*/

      struct simple_counters: public abstract_stats
      {
	simple_counters();

	void		operator () (unsigned	type,
				     unsigned	subtype,
				     bool	retransmitted);
	std::ostream&	print(std::ostream& o) const;
	unsigned	total() const;

      private:
	unsigned	total_;
	unsigned	retransmitted_;
	unsigned	type_count_[4];
	unsigned	subtype_count_[4][16];
      };

      /*------------.
      | gap_lengths |
      `------------*/

      struct gap_lengths: public abstract_stats
      {
	gap_lengths();

	void			operator () (unsigned gap_length);
	std::ostream&		print(std::ostream& o) const;

      private:
	typedef std::vector<unsigned>		gaplen_frequencies;

	enum
	  {
	    gl_max		= 2048,
	    gl_freqs_resolution	= 16,
	    gl_freqs_size	= gl_max / gl_freqs_resolution
	  };

	gaplen_frequencies	gl_freqs_;
	unsigned		gap_count_;
      };

      /*-------------------.
      | missed_estimations |
      `-------------------*/

      struct missed_estimations: public abstract_stats
      {
	missed_estimations(const simple_counters& c);

	void		operator () ();
	void		operator () (unsigned gap_length);
	std::ostream&	print(std::ostream& o) const;

      private:
	enum
	  {
	    threshold_steps	= 16,
	    missed_size		= 2048 / threshold_steps
	  };

	unsigned		missed_[missed_size];
	const simple_counters*	c_;
      };

      /*------.
      | stats |
      `------*/

      struct stats: public abstract_stats
      {
	stats();

	template <class PhyHeader, class Frame>
	void
	account(const Frame& f);

	void				clear_states();
	std::ostream&			print(std::ostream& o) const;

	const simple_counters&		counters() const;
	const gap_lengths&		gaplens() const;
	const missed_estimations&	missed() const;

      private:
	enum
	  {
	    state_lifetime = 60 * 1000 * 1000 // 60 second.
	  };

	struct state
	{
	  state(unsigned			last_frame_id,
		const seqctl&			last_sc,
		const tool::microseconds&	last_rx_stamp);

	  unsigned		last_frame_id;
	  seqctl		last_sc;
	  tool::microseconds	last_rx_stamp;
	};

	typedef boost::tuple<addr, addr, uint8_t>	state_id;
	typedef std::map<state_id, state>		states;

	struct hooks: public dissector_default_hooks
	{
	  hooks();

	  void	addr_hook(const void*, size_t, unsigned, const addr&);
	  void	seq_ctl_hook(const void*, size_t, unsigned, unsigned);
	  void	qos_ctl_hook(const void*, size_t,
			     unsigned, bool, unsigned, unsigned);

	  typedef boost::optional<seqctl>	optional_seqctl;
	  typedef boost::optional<state_id>	optional_state_id;

	  unsigned		type;
	  unsigned		subtype;
	  bool			retx;
	  optional_seqctl	sc;
	  optional_state_id	id;

	private:
	  const addr*		receiver;
	};

      private:
	states			states_;

	simple_counters		counters_;
	gap_lengths		gap_lengths_;
	missed_estimations	missed_;
      };

      std::ostream&
      operator << (std::ostream& os, const abstract_stats& s);

    } // End of namespace wifi::frame::stats.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

# include "stats.hxx"

#endif // ! WIFI_FRAME_STATS_HH_
