/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_STATS_HXX_
# define WIFI_FRAME_STATS_HXX_

# include <wipal/wifi/frame/stats.hh>

# include <wipal/wifi/frame/dissector/dissector.hh>

namespace wifi
{

  namespace frame
  {

    namespace stats
    {

      /*---------------.
      | abstract_stats |
      `---------------*/

      inline
      abstract_stats::~abstract_stats()
      {
      }

      /*----------------.
      | simple_counters |
      `----------------*/

      inline
      simple_counters::simple_counters(): total_ (0),
					  retransmitted_ (0)
      {
	memset(type_count_, 0, sizeof (type_count_));
	memset(subtype_count_, 0, sizeof (subtype_count_));
      }

      inline
      void
      simple_counters::operator () (unsigned	type,
				    unsigned	subtype,
				    bool	retransmitted)
      {
	assert(type < 4);
	assert(subtype < 16);

	++total_;
	if (retransmitted)
	  ++retransmitted_;
	++type_count_[type];
	++subtype_count_[type][subtype];
      }

      inline
      std::ostream&
      simple_counters::print(std::ostream& o) const
      {
	o << "frame count: " << total_ << '\n'
	  << "retransmissions: " << retransmitted_ << '\n'
	  << "retransmissions: " << (100.l * retransmitted_ / total_) << "%\n"
	     "\n"
	     "begin frame types/subtypes\n";
	for (unsigned t = 0; t < 4; ++t)
	  {
	    const unsigned	tc = type_count_[t];

	    if (not tc)
	      continue;

	    o << "  " << type::names[t] << ": " << (100.l * tc / total_) << "%"
	         " (" << tc << ")\n";

	    for (unsigned st = 0; st < 16; ++st)
	      {
		const unsigned	stc = subtype_count_[t][st];

		if (not stc)
		  continue;

		o << "\t" << subtype::names[t][st] << ": "
		  << (100.l * stc / total_) << "% of total, "
		  << (100.l * stc / tc) << "% of this type"
		     " (" << stc << ")\n";
	      }
	  }
	o << "end frame types/subtypes" << std::endl;
	return o;
      }

      inline
      unsigned
      simple_counters::total() const
      {
	return total_;
      }

      /*------------.
      | gap_lengths |
      `------------*/

      inline
      gap_lengths::gap_lengths(): gl_freqs_ (gl_freqs_size),
				  gap_count_ (0)
      {
      }

      inline
      void
      gap_lengths::operator () (unsigned gap_length)
      {
	assert(1 < gap_length);
	assert(gap_length < gl_max);

	++gl_freqs_[gap_length / gl_freqs_resolution];
	++gap_count_;
      }

      inline
      std::ostream&
      gap_lengths::print(std::ostream& o) const
      {
	for (unsigned i = 0; i < gl_freqs_.size(); ++i)
	  {
	    const unsigned	f = gl_freqs_[i];

	    if (not f)
	      continue;

	    o << "  " << (i * gl_freqs_resolution + (gl_freqs_resolution / 2))
	      << '\t' << (double (f) / gap_count_) << '\n';
	  }
	return o << std::flush;
      }

      /*-------------------.
      | missed_estimations |
      `-------------------*/

      inline
      missed_estimations::missed_estimations(const simple_counters& c): c_ (&c)
      {
	memset(missed_, 0, sizeof (missed_));
      }

      inline
      void
      missed_estimations::operator () ()
      {
	for (unsigned i = 0; i < sizeof (missed_) / sizeof (*missed_); ++i)
	  ++missed_[i];
      }

      inline
      void
      missed_estimations::operator () (unsigned gap_length)
      {
	for (unsigned i = 0; i < sizeof (missed_) / sizeof (*missed_); ++i)
	  if (int (gap_length) < (int (i) + 1) * threshold_steps)
	    missed_[i] += gap_length - 1;
      }

      inline
      std::ostream&
      missed_estimations::print(std::ostream& o) const
      {
	for (unsigned i = 0; i < sizeof (missed_) / sizeof (*missed_); ++i)
	  o << "  " << ((int (i) + 1) * threshold_steps) << '\t'
	    << (double (missed_[i]) / (missed_[i] + c_->total())) << '\n';
	return o << std::flush;
      }

      /*------.
      | stats |
      `------*/

      /*---------------------.
      | Various constructors |
      `---------------------*/

      inline
      stats::stats(): missed_ (counters_)
      {
      }

      inline
      stats::hooks::hooks(): type (-1),
			     subtype (-1),
			     retx (false),
			     receiver (0)
      {
      }


      inline
      stats::state::state(unsigned			last_frame_id,
			  const seqctl&			last_sc,
			  const tool::microseconds&	last_rx_stamp):
	last_frame_id (last_frame_id),
	last_sc (last_sc),
	last_rx_stamp (last_rx_stamp)
      {
      }

      /*-------.
      | Misc.  |
      `-------*/

      inline
      void
      stats::clear_states()
      {
	states_.clear();
      }

      inline
      std::ostream&
      stats::print(std::ostream& o) const
      {
	return o << counters()
		 << "\n"
	            "begin gap length frequencies\n"
		 << gaplens()
		 << "end gap length frequencies\n"
	            "\n"
		    "begin missed frames estimations\n"
		 << missed()
		 << "end missed frames estimation" << std::endl;
      }

      inline
      const simple_counters&
      stats::counters() const
      {
	return counters_;
      }

      inline
      const gap_lengths&
      stats::gaplens() const
      {
	return gap_lengths_;
      }

      inline
      const missed_estimations&
      stats::missed() const
      {
	return missed_;
      }

      /*-----------------------------.
      | Frame parsing and accounting |
      `-----------------------------*/

      inline
      void
      stats::hooks::addr_hook(const void* f, size_t,
			      unsigned i, const addr& a)
      {
	switch (i)
	  {
	  case 1:
	    type = type_of(f);
	    subtype = subtype_of(f);
	    retx = control_flag(3, f);
	    receiver = &a;
	    break;

	  case 2:
	    if (type == frame::type::management or type == frame::type::data)
	      id = state_id (a, addr::broadcast());
	    break;
	  }
      }

      inline
      void
      stats::hooks::seq_ctl_hook(const void*, size_t, unsigned fn, unsigned sn)
      {
	sc = seqctl (sn, fn);
      }

      inline
      void
      stats::hooks::qos_ctl_hook(const void*, size_t,
				 unsigned tid,
				 bool, unsigned, unsigned)
      {
	assert(id);

	if (*receiver == addr::broadcast())
	  return;

	id->get<1>() = *receiver;
	id->get<2>() = tid;
      }

      template <class PhyHeader, class Frame>
      void
      stats::account(const Frame& f)
      {
	const dissector<hooks>	d = dissect<PhyHeader, hooks>(f);

	if (d.status() == dissector_status::invalid)
	  return;

	counters_(d.type, d.subtype, d.retx);

	if (not d.id or not d.sc) // No state to update.
	  return;

	const unsigned			frame_id = f.pcap_header()->id;
	const tool::microseconds&	rx_stamp = f.microseconds();
	const states::iterator		i = states_.find(*d.id);

	state				new_state (frame_id, *d.sc, rx_stamp);

	// First occurence of this state ID.
	if (i == states_.end())
	  {
	    states_.insert(std::make_pair(*d.id, new_state));
	    return;
	  }

	const state&		s = i->second;

	// Check the previous state has not expired.
	if (rx_stamp - s.last_rx_stamp < state_lifetime)
	  {
	    // Ignore out-of-sequence retransmitted frames.
	    if (*d.sc < s.last_sc and d.retx)
	      return;

	    // We missed the original non-retransmitted frame.
	    if (*d.sc > s.last_sc and d.retx)
	      missed_();

	    const int	delta = d.sc->seqnum_get() - s.last_sc.seqnum_get();

	    // There is a sequence number gap.
	    if (1 < delta)
	      {
		gap_lengths_(delta);
		missed_(delta);
	      }
	  }
	i->second = new_state;
      }

      /*------------.
      | operator << |
      `------------*/

      inline
      std::ostream&
      operator << (std::ostream& os, const abstract_stats& s)
      {
	return s.print(os);
      }

    } // End of namespace wifi::frame::stats.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_STATS_HXX_
